/* 
 * i-scream central monitoring system
 * http://www.i-scream.org.uk
 * Copyright (C) 2000-2002 i-scream
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

//---PACKAGE DECLARATION---
package uk.org.iscream.cms.corbaservices;

//---IMPORTS---
import java.lang.reflect.*;
import java.util.*;
import java.io.*;
import java.net.*;

/**
 * CORBA Services Manager
 *
 * This class provides a simple wrapper to CORBA services
 * that require a web service to "share" their IOR files.
 * This provides an all-in-one way of starting up service(s)
 * and serving their IOR files. No external web servers are
 * required.
 *
 * @author  $Author: tdb $
 * @version $Id: CorbaServices.java,v 1.12 2003/02/21 13:45:47 tdb Exp $
 */
class CorbaServices {

//---FINAL ATTRIBUTES---

    /**
     * The current CVS revision of this class
     */
    public static final String REVISION = "$Revision: 1.12 $";
    
    /**
     * Default path to "ok URL's" list
     */
    public static final String DEFAULTOKURLS = "./etc/okUrls.conf";
    
    /**
     * Default path to services list
     */
    public static final String DEFAULTSERVICESCONF = "./etc/services.conf";
    
    /**
     * Default path to web directory. Note there is *no* trailing slash.
     */
    public static final String DEFAULTWEBDIR = "./web";
    
    /**
     * Default port number
     */
    public static final int DEFAULTPORT = 8052;

    /**
     * The default location of the properties file for the system
     */
    public static final String DEFAULTPROPERTIES = "./etc/default.properties";
    
//---STATIC METHODS---
    
    /**
     * This main method starts the whole system up. A lot
     * of the functionality is provided in this main class.
     */
    public static void main(String args[]) {
        
        System.out.println("-----------------------------------------");
        System.out.println("---  i-scream CORBA Services Manager  ---");
        System.out.println("---   (c) 2001 The i-scream Project   ---");
        System.out.println("---    (http://www.i-scream.org.uk)   ---");
        System.out.println("-----------------------------------------");
        System.out.println("---          Starting System          ---");
        
        // Read command line arguments
        String okUrls = DEFAULTOKURLS;
        String servicesConf = DEFAULTSERVICESCONF;
        String webDir = DEFAULTWEBDIR;
        int port = DEFAULTPORT;
        String defaultProperties = DEFAULTPROPERTIES;
        for(int i=0; i < args.length; i++) {
            if(args[i].equals("-h")) {
                usage();
            }
            else if(args[i].equals("-u")) {
                if(++i < args.length) {
                    okUrls = args[i];
                }
                else {
                    usage();
                }
            }
            else if(args[i].equals("-s")) {
                if(++i < args.length) {
                    servicesConf = args[i];
                }
                else {
                    usage();
                }
            }
            else if(args[i].equals("-w")) {
                if(++i < args.length) {
                    webDir = args[i];
                }
                else {
                    usage();
                }
            }
            else if(args[i].equals("-p")) {
                if(++i < args.length) {
                    port = Integer.parseInt(args[i]);
                }
                else {
                    usage();
                }
            }
            else if(args[i].equals("-l")) {
                if(++i < args.length) {
                    defaultProperties = args[i];
                }
                else {
                    usage();
                }
            }
            else {
                usage();
            }
        }
        
        // load the default properties file into the system properties
        System.out.println("initialising: reading " + defaultProperties);
        try {
            Properties initProperties = new Properties(System.getProperties());
            initProperties.load(new FileInputStream(new File(defaultProperties))
);
            System.setProperties(initProperties);
        } catch (Exception e) {
            System.err.println("ERROR: " + e.getMessage());
            usage();
        }
        
        // Read in the services configuration
        System.out.println("---       Reading Configuration       ---");
        
        LinkedList serviceList = getServicesList(servicesConf);
        String[] svcList = new String[serviceList.size()];
        LinkedList[] svcListParams = new LinkedList[serviceList.size()];
        Iterator i = serviceList.iterator();
        int j=0;
        while(i.hasNext()) {
            StringTokenizer st = new StringTokenizer((String)i.next());
            svcList[j] = (String) st.nextToken();
            LinkedList l = new LinkedList();
            while (st.hasMoreTokens()) {
                l.add(st.nextToken());
            }
            svcListParams[j] = l;
            j++;
        }
        
        // Start the services one by one
        System.out.println("---      Starting CORBA Services      ---");
        
        try {
            
            // get hooks on the system loaders
            Runtime myRuntime = Runtime.getRuntime();
            ClassLoader myClassLoader = ClassLoader.getSystemClassLoader();
            
            // start each service one by one
            for (int k=0; k < svcList.length; k++) {
                
                // Load the class
                Class myClass = myClassLoader.loadClass(svcList[k]);
                // Sort out the command line parameters
                String[] passedArgs = new String[svcListParams[k].size()];
                // keep a String list for printing out later :)
                String arglist = "";
                i = svcListParams[k].iterator();
                int p = 0;
                while(i.hasNext()) {
                    passedArgs[p] = (String) i.next();
                    arglist += " " + passedArgs[p];
                    p++;
                }
                                                        
                // create an array of classes that are listed in the main method header
                // so we can get a hook to the main method
                Class[] passedArgsClass = {passedArgs.getClass()};
                // get a hook on the main method
                Method myMethod = myClass.getMethod("main",passedArgsClass);
                
                // Now we know where it is, we need to create a thread for it, passing in 
                // a list of objects that represent its args.
                Object[] passedArgsObject = {passedArgs};               
                ServiceThread aSvc = new ServiceThread(myMethod, passedArgsObject);
    
                // Finally we start it going
                aSvc.start();
                System.out.println(svcList[k]+arglist);        
            }
        
        } catch (Exception e) {
            System.out.println(e);
            e.printStackTrace();
        }
        
        System.out.println("---      Starting Mini WebServer      ---");
        
        // print out some system details
        String host;
        try {
            host = InetAddress.getLocalHost().getHostName();
        } catch(UnknownHostException e) {
            host = "<unknown>";
        }
        System.out.println("Host:\t\t"+host);
        System.out.println("Port:\t\t"+port);
        System.out.println("URL List File:\t"+okUrls);
        System.out.println("Web Directory:\t"+webDir);
        
        // start the mini-webserver
        MiniWebServer mws = new MiniWebServer(port, okUrls, webDir);
        mws.start();

        System.out.println("---   i-scream CORBA Services Ready   ---");
    }
    
    /**
     * Print out usage information.
     */
    private static void usage() {
        System.out.println("USAGE: java uk.org.iscream.cms.corbaservices.CorbaServices <option>");
        System.out.println("   or: java -jar iscream-corbaservices.jar <option>");
        System.out.println("WHERE <option>:");
        System.out.println("      -u <filename> - the location of okUrls config file");
        System.out.println("                      the default is ./etc/okUrls.conf");
        System.out.println("      -s <filename> - the location of services config file");
        System.out.println("                      the default is ./etc/services.conf");
        System.out.println("      -w <dir>      - the location of the web directory");
        System.out.println("                      the default is ./web  (no trailing /)");
        System.out.println("      -p <port>     - this port to bind the webserver to");
        System.out.println("                      the default is 8052");
        System.out.println("      -l <filename> - the location of initial system properties");
        System.out.println("                      the default is ./etc/default.properties");
        System.out.println("      -h            - this help screen");
        System.exit(1);
    }
    
    /**
     * Read the services into a LinkedList. This is later
     * parsed to get classes and parameters.
     *
     * @param filename the services configuration file
     * @return a LinkedList of services
     */
    private static LinkedList getServicesList(String filename) {
        File file = new File(filename);
        LinkedList svcList = new LinkedList();
        if(file.exists() && file.canRead()) {
            // if we can read the file, do so into svcList
            try {
                BufferedReader reader = new BufferedReader(new FileReader(file));
                while(reader.ready()) {
                    String line = reader.readLine();
                    // ignore comments
                    if(!line.startsWith("#")) {
                        svcList.add(line);
                    }
                }
            } catch(IOException e) {
                System.out.println(e);
                e.printStackTrace();
            }
        }
        return svcList;
    }

//---CONSTRUCTORS---

//---PUBLIC METHODS---

//---PRIVATE METHODS---

//---ACCESSOR/MUTATOR METHODS---

//---ATTRIBUTES---

//---STATIC ATTRIBUTES---

//---INNER CLASSES---
    
    /**
     * Inner class to provide a thread for running a Service.
     */
    private static class ServiceThread extends Thread {
     
        /**
         * The constructor takes the name of the method we're gonna run
         * and the arguments we're gonna pass into it.
         *
         * @param mainMethod a method that we'll treat as main
         * @param args an array of arguments for mainMethod
         */
        ServiceThread(Method mainMethod, Object[] args) {
            _mainMethod = mainMethod;
            _args = args;
        }
        
        /**
         * Invokes our main method, with the arguments we
         * have been given.
         */
        public void run() {
            try {
                _mainMethod.invoke(null, _args);
            } catch (Exception e){
                System.out.println(e);
                e.printStackTrace();
            }
        }
        
        /**
         * The method we'll treat as main
         */
        Method _mainMethod = null;
        
        /**
         * The array of arguments
         */
        Object[] _args = null;
    }

}
